{
  Special memory manager to hard check memory heap corruptions.
  To use it, just place a reference to this unit *FIRST* in the uses clause
  of your project (dpr-file).
  =====================================================================
  Copyright (C) by Vladimir Kladov, 2007
  ---------------------------------------------------------------------
  http://kolmck.net
  mailto: vk@kolmck.net
}

unit MMHardCheck;

interface

uses windows;

procedure HardCheck;

implementation

{$DEFINE SELFTEST}

const MMSignature0 = $CE8A4602;
      MMSignature1 = $DF9B5713;

type
  PMMSig = ^TMMSig;
  TMMSig = packed record
    Signature0: DWORD;
    Size: Integer;
    PrevChain: PMMSig;
    NextChain: PMMSig;
    Signature1: DWORD;
  end;

var OldMM: TMemoryManager;
    MMSem: THandle;
    ChkOn: Boolean = TRUE; // set this value to FALSE to stop constant hard checking of heap!
    LastChain: PMMSig;
    CountAlloc: Integer;

procedure HardCheck;
asm
  mov eax, [LastChain]
  mov ecx, [CountAlloc]
  dec ecx
  jle @@1

@@loop_chk_1:
  mov edx, [eax].TMMSig.Signature0
  cmp edx, MMSignature0
  jz  @@sig0_ok
  int 3 // !
@@sig0_ok:
  mov edx, [eax].TMMSig.Signature1
  cmp edx, MMSignature1
  jz  @@sig1_ok
  int 3 // !
@@sig1_ok:
  mov eax, [eax].TMMSig.NextChain
  loop @@loop_chk_1

@@1:
  cmp ecx, 0
  jl  @@exit

  mov edx, [eax].TMMSig.Signature0
  cmp edx, MMSignature0
  jz  @@sig0_ok2
  int 3 // !
@@sig0_ok2:
  mov edx, [eax].TMMSig.Signature1
  cmp edx, MMSignature1
  jz  @@sig1_ok2
  int 3 // !
@@sig1_ok2:
  mov ecx, [eax].TMMSig.NextChain
  JECXZ @@exit
  int 3 // !

@@exit:
end;

function HardGetMem(size: Integer): Pointer;
// Allocate memory block.
var S: PMMSig;
begin
  if IsMultiThread and (MMSem <> 0) then
    WaitForSingleObject( MMSem, INFINITE );
  TRY
    if ChkOn then
      HardCheck;
    S := OldMM.GetMem( size + Sizeof( TMMSig ) );
    S.Size := size;
    S.PrevChain := nil;
    S.NextChain := LastChain;
    S.Signature0 := MMSignature0;
    S.Signature1 := MMSignature1;
    if LastChain <> nil then
      LastChain.PrevChain := S;
    LastChain := S;
    Result := Pointer( DWORD( S ) + Sizeof( TMMSig ) );
    inc( CountAlloc );
  FINALLY
    {$IFDEF SELFTEST}
    if ChkOn then
      HardCheck;
    {$ENDIF}
    if IsMultiThread and (MMSem <> 0) then
      ReleaseSemaphore( MMSem, 1, nil );
  END;
end;

function HardFreeMem(p: Pointer): Integer;
// Deallocate memory block.
var S: PMMSig;
begin
  if IsMultiThread and (MMSem <> 0) then
    WaitForSingleObject( MMSem, INFINITE );
  TRY
    S := Pointer( DWORD( p ) - Sizeof( TMMSig ) );
    if (S.Signature0 <> MMSignature0) or
       (S.Signature1 <> MMSignature1) then
    asm
      int 3 //   ,  :
            //      !
    end;
    if ChkOn then
      HardCheck;
    if S.NextChain <> nil then
      S.NextChain.PrevChain := S.PrevChain;
    if S.PrevChain <> nil then
      S.PrevChain.NextChain := S.NextChain
    else
    begin
      if S <> LastChain then
      asm
        int 3 //   !
      end
        else
      begin
        LastChain := S.NextChain;
      end;
    end;
    S.Signature0 := $CCCCCCCC;
    S.Signature1 := $DDDDDDDD;
    Result := OldMM.FreeMem( S );
    Dec( CountAlloc );
  FINALLY
    {$IFDEF SELFTEST}
    if ChkOn then
      HardCheck;
    {$ENDIF}
    if IsMultiThread and (MMSem <> 0) then
      ReleaseSemaphore( MMSem, 1, nil );
  END;
end;

function min( a, b: Integer ): Integer;
asm
  cmp EAX, EDX
  jle @@1
  xchg EAX, EDX
@@1:
end;

function HardReallocMem(p: Pointer; size: Integer): Pointer;
// Resize memory block.
var S, Sold: PMMSig;
begin
  if IsMultiThread and (MMSem <> 0) then
    WaitForSingleObject( MMSem, INFINITE );
  TRY
    if ChkOn then
      HardCheck;
    Sold := Pointer( DWORD( p ) - Sizeof( TMMSig ) );
    if (Sold.Signature0 <> MMSignature0) or
       (Sold.Signature1 <> MMSignature1) then
    asm
      int 3 //   ,  :
            //      !
    end;
    if Sold.Size >= size then
    begin
      Result := p;
      Exit;
    end;
    S := OldMM.GetMem( size + Sizeof( TMMSig ) );
    S.Size := size;
    S.PrevChain := nil;
    S.NextChain := LastChain;
    S.Signature0 := MMSignature0;
    S.Signature1 := MMSignature1;
    if LastChain <> nil then
      LastChain.PrevChain := S;
    LastChain := S;
    Result := Pointer( DWORD( S ) + Sizeof( TMMSig ) );
    move( p^, Result^, min( Sold.Size, size ) );
    //Inc( CountAlloc );

    S := Sold;
    if S.NextChain <> nil then
      S.NextChain.PrevChain := S.PrevChain;
    if S.PrevChain <> nil then
      S.PrevChain.NextChain := S.NextChain
    else
    begin
      if S <> LastChain then
      asm
        int 3 //   !
      end
        else
      begin
        LastChain := S.NextChain;
      end;
    end;
    S.Signature0 := $CCCCCCCC;
    S.Signature1 := $DDDDDDDD;
    OldMM.FreeMem( S );
    //Dec( CountAlloc );
  FINALLY
    {$IFDEF SELFTEST}
    if ChkOn then
      HardCheck;
    {$ENDIF}
    if IsMultiThread and (MMSem <> 0) then
      ReleaseSemaphore( MMSem, 1, nil );
  END;
end;

const
  HardMemoryManager: TMemoryManager = (
    GetMem: HardGetMem;
    FreeMem: HardFreeMem;
    ReallocMem: HardReallocMem);

initialization

  MMSem := CreateSemaphore( nil, 1, 1, nil );

  GetMemoryManager( OldMM );
  SetMemoryManager( HardMemoryManager );

finalization

  //SetMemoryManager( OldMM );
  CloseHandle( MMSem );
  MMSem := 0;


end.
